#!/bin/env python
#
#    ldab.py
#    Latent Dirichlet allocation with a background distribution.
#    $Id: ldab.py,v 1.2 2020/08/02 15:34:46 daichi Exp $
#
import ldac
import re
import sys
import gzip
import getopt
import pickle
import numpy as np
import numpy.random as npr
import loggings
from   loggings import elprintf
from   scipy.special import gammaln
from   collections import defaultdict

# global parameters
K     = 10
iters = 1
alpha = 0
beta  = 0.01
eta   = 0.005
a0    = 2.0
a1    = 1.0
threshold = 10

class stats:
    def __init__ (self, K, D, V):
        self.nz  = np.zeros (K, dtype=np.int32)
        self.ndz = np.zeros ((D,K), dtype=np.int32)
        self.nwz = np.zeros ((V,K), dtype=np.int32)
        # for background distribution
        self.nb  = np.zeros (D, dtype=np.int32)
        self.nw  = np.zeros (V, dtype=np.int32)
        self.nws = np.zeros (1, dtype=np.int32)

# main functions.

def textread (file, lexicon):
    docs = []
    with open (file, 'r') as fh:
        doc = []
        for line in fh:
            if re.match (r'^[ \t]*$', line):
                if len(doc) > 0:
                    docs.append (doc)
                    doc = []
            else:
                words = line.rstrip('\n').split()
                for word in words:
                    if word in lexicon:
                        id = lexicon[word]
                        doc.append (id)
        if len(doc) > 0:
            docs.append (doc)
    return docs

def lexread (file, threshold):
    freq = defaultdict (int)
    id = 0
    lexicon = {}
    with open (file, 'r') as fh:
        for line in fh:
            if re.match (r'^[ \t]*$', line):
                continue
            words = line.rstrip('\n').split()
            for word in words:
                freq[word] += 1
    # create lexicon
    for word,cnt in sorted (freq.items(), key=lambda x: x[1], reverse=True):
        if cnt >= threshold:
            lexicon[word] = id
            id += 1
    return lexicon
            
def ldaread (file, K, lexicon):
    docs = textread (file, lexicon)
    topics = randtopic (docs, K)
    return int32(docs), int32(topics)
    
def randtopic (words, K):
    p = 0.5 # ratio of background
    topics = []
    for word in words:
        base = npr.binomial (1,p,size=len(word))
        topic = base * (1 + npr.randint(K, size=len(word)))
        topics.append (topic)
    return topics

def ldappl (docs,zz,stat,alpha,beta,eta,a0,a1):
    L = datalen (docs)
    lik = polyap (zz,a0,a1) + polyab (stat.nw,eta) + \
          polyad (stat.ndz,alpha) + polyaw (stat.nwz,beta)
    return np.exp (- lik / L)

def polyab (NW,eta):
    V = len(NW)
    lik = gammaln (V * eta) - gammaln (V * eta + np.sum(NW)) \
          + np.sum (gammaln (NW + eta) - gammaln (eta))
    return lik

def polyap (zz,a0,a1):
    D = len(zz)
    lik = 0
    for n in range(D):
        L = len (zz[n])
        n0 = sum (zz[n] == 0)
        n1 = L - n0
        lik += gammaln (a0 + a1) - gammaln (L + a0 + a1) + \
               gammaln (a0 + n0) - gammaln (a0) + \
               gammaln (a1 + n1) - gammaln (a1)
    return lik

def polyad (ND,alpha):
    D = ND.shape[0]
    K = ND.shape[1]
    nd = np.sum(ND,1)
    lik = np.sum (gammaln (K * alpha) - gammaln (K * alpha + nd))
    for n in range(D):
        lik += np.sum (gammaln (ND[n,:] + alpha) - gammaln (alpha))
    return lik

def polyaw (NW,beta):
    V = NW.shape[0]
    K = NW.shape[1]
    nw = np.sum(NW,0)
    lik = np.sum (gammaln (V * beta) - gammaln (V * beta + nw))
    for k in range(K):
        lik += np.sum (gammaln (NW[:,k] + beta) - gammaln (beta))
    return lik

def stopvec (file, lexicon, V):
    if not lexicon:
        return None
    eprint('using stopwords from %s.' % file)
    # body
    vec = np.zeros (V,dtype=np.int32)
    with open (file, 'r') as fh:
        for line in fh:
            word = line.rstrip('\n')
            if word in lexicon:
                id = lexicon[word]
                vec[id] = 1
    return vec

# supporting functions.
    
def nlex (docs):
    v = 0
    for doc in docs:
        if max(doc) > v:
            v = max(doc)
    return v + 1

def int32 (docs):
    data = []
    for doc in docs:
        data.append (np.array(doc, dtype=np.int32))
    return data

def datalen (docs):
    n = 0
    for doc in docs:
        n += len (doc)
    return n

def eprint (s):
    sys.stderr.write (s + '\n')
    sys.stderr.flush ()

def eprintf (s):
    sys.stderr.write (s)
    sys.stderr.flush ()

def datalen (W):
    return sum (map (lambda x: x.shape[0], W))

def normalize (x):
    return x / np.sum(x)

def cnormalize (M): # column-wise normalize matrix
    s = np.sum(M,0)
    return np.dot(M,np.diag(1.0/s))

def rnormalize (M): # row-wise normalize matrix
    return np.array ([m / np.sum(m) for m in M])
    
def bgrate (zzs):
    N = 0.0; b = 0.0
    for zz in zzs:
        N += len(zz)
        for z in zz:
            if z == 0:
                b += 1
    return (b / N * 100)

def show (file, zz, docs, lexicon):
    with open (file, 'w') as fh:
        # prepare dictionary
        idword = {}
        for word,id in lexicon.items():
            idword[id] = word
        # print words
        for n in range(len(zz)):
            fh.write ('[document %d]\n' % n)
            for i in range(len(zz[n])):
                w = docs[n][i]
                z = zz[n][i]
                fh.write (' %-10s\t%d\n' % (idword[w], z))

# I/O functions.
        
def save (file,lexicon,stop,words,zz,stat,alpha,beta,eta):
    eprint ('saving model to %s .. ' % file)
    model = {}
    model['alpha'] = alpha
    model['eta']   = eta
    model['beta']  = cnormalize (stat.nwz + beta)
    model['theta'] = rnormalize (stat.ndz + alpha)
    model['background'] = normalize (stat.nw + eta)
    if lexicon:
        model['lexicon'] = lexicon
    if stop is not None:
        model['stopwords'] = stop
    # save to file
    with gzip.open (file, 'wb') as gf:
        pickle.dump (model, gf, 2)
    eprint ('done.')

def savewz (file, W, Z):
    D = len (W)
    with gzip.open (file + 'wz.gz', 'w') as fh:
        for n in range(D):
            N = W[n].shape[0]
            for i in range(N):
                fh.write('%d\t%d\n' % (W[n][i], Z[n][i]))
            fh.write('\n')

def parse_args (argv):
    global K,iters,threshold,alpha,beta,eta
    shortopts = "K:N:t:a:b:e:h"
    longopts = ['topics=','iters=','threshold=','alpha=','beta=','eta=','help']
    # parse argv
    try:
        opts, args = getopt.getopt (argv, shortopts, longopts)
    except getopt.GetoptError as err:
        usage ()
        
    for o,a in opts:
        if o in ('-K', '--topics'):
            K     = int (a)
        elif o in ('-N', '--iters'):
            iters = int (a)
        elif o in ('-t', '--threshold'):
            threshold = int (a)
        elif o in ('-a', '--alpha'):
            alpha = float (a)
        elif o in ('-b', '--beta'):
            beta = float (a)
        elif o in ('-e', '--eta'):
            eta = float (a)
        elif o in ('-h', '--help'):
            usage ()
        else:
            assert False, "unknown option"
    if alpha == 0:
        alpha = 50.0 / K

    if len(args) >= 2:
        train     = args[0]
        model     = args[1]
        stopwords = args[2] if len(args) >= 3 else None
        return train, model, stopwords
    else:
        usage ()

def usage ():
    print ('LDAb: LDA with a background distribution.')
    print ('$Id: ldab.py,v 1.2 2020/08/02 15:34:46 daichi Exp $')
    print ('usage: % ldab.py OPTIONS train.txt model [stopwords]')
    print ('OPTIONS')
    print (' -K topics  number of topics in LDA')
    print (' -N iters   number of Gibbs iterations (default 1)')
    print (' -t freq    threshold of word frequency (default 10)')
    print (' -a alpha   Dirichlet hyperparameter on topics (default 50/K)')
    print (' -b beta    Dirichlet hyperparameter on words (default 0.01)')
    print (' -e eta     Dirichlet hyperparameter on background (default 0.005)')
    print (' -h         displays this help')
    sys.exit (0)

def main ():
    global K,iters,alpha,beta,eta
    train,model,stopwords = parse_args (sys.argv[1:])

    eprint ('LDAb: K=%d, iters=%d, threshold=%d, alpha=%g, beta=%g, eta=%g' \
            % (K,iters,threshold,alpha,beta,eta))
            
    loggings.start (sys.argv, model)
    
    lexicon = lexread (train, threshold)
    docs,zz = ldaread (train, K, lexicon)
    D  = len (docs)
    V  = nlex (docs)
    N  = datalen (docs)
    stat = stats (K, D, V)
    stop = stopvec (stopwords, lexicon, V) if stopwords else None
    eprint ('documents = %d, lexicon = %d, nwords = %d' % (D,V,N))

    for iter in range(iters):
        ldac.gibbs (docs, zz, stat, stop, alpha, beta, eta, a0, a1, iter)
        elprintf('Gibbs iteration [%2d/%2d] PPL = %.2f (bg=%.1f%%)' \
                 % (iter+1, iters,
                    ldappl (docs,zz,stat,alpha,beta,eta,a0,a1),
                    bgrate(zz)))
    eprintf('\n')

    save (model,lexicon,stop,docs,zz,stat,alpha,beta,eta)
    
    loggings.finish ()

if __name__ == "__main__":
    main ()

