#!/usr/local/bin/python

import numpy as np
import numpy.random as npr
cimport numpy as np
cimport cython
ctypedef np.int32_t dtype_t
ctypedef np.float64_t ftype_t

@cython.boundscheck(False)

def init (list words, list zz, stat):
    D = len(words)
    for d in xrange(D):
        N = len(zz[d])
        for n in xrange(N):
            w = words[d][n]
            z = zz[d][n]
            if z == 0:
                stat.nb[d]  += 1
                # global statistics
                stat.nw[w]  += 1
                stat.nws[0] += 1
            else:
                z = z - 1
                stat.ndz[d,z] += 1
                stat.nwz[w,z] += 1
                stat.nz[z]    += 1

def gibbs (list words, list zz, stat, np.ndarray[dtype_t,ndim=1] stopwords,
           double alpha, double beta, double eta, double a0, double a1, int iter):
    # alias
    cdef np.ndarray[dtype_t,ndim=1] nz  = stat.nz
    cdef np.ndarray[dtype_t,ndim=1] nb  = stat.nb
    cdef np.ndarray[dtype_t,ndim=1] nw  = stat.nw
    cdef np.ndarray[dtype_t,ndim=1] nws = stat.nws
    cdef np.ndarray[dtype_t,ndim=2] ndz = stat.ndz
    cdef np.ndarray[dtype_t,ndim=2] nwz = stat.nwz
    
    # constants, arrays
    cdef int D = ndz.shape[0]
    cdef int K = ndz.shape[1]
    cdef int V = nwz.shape[0]
    cdef np.ndarray[ftype_t,ndim=1] p = np.zeros (K, dtype=np.float64)
    cdef np.ndarray[dtype_t,ndim=1] seq = np.zeros (D, dtype=np.int32)
    cdef np.ndarray[ftype_t,ndim=1] rands
    # variables
    cdef int j, k, d, n, w, z, N
    cdef double s, t, r, q, p0, p1

    # prepare
    for n in range(D):
        seq[n] = n
    npr.shuffle (seq)

    # sample
    for j in range(D):
        d = seq[j]
        N = len (zz[d])
        rands = npr.rand (2 * N)
        for n in range(N):
            w = words[d][n]
            z = zz[d][n]
            s = 0.0

            # decrement
            if (iter > 0):
                if z == 0:
                    nb[d]    -= 1
                    # global statistics
                    nw[w]    -= 1
                    nws[0]   -= 1
                else:
                    z = z - 1
                    ndz[d,z] -= 1
                    nwz[w,z] -= 1
                    nz[z]    -= 1

            # compute p(x = 0)
            if (iter > 0):
                q = (a0 + nb[d]) / (a0 + a1 + N - 1)
            else:
                q = 0.5

            # p(z = k,w|x = 1) = p(w|z = k)p(z = k)
            for k in range(K):
                p[k] = (ndz[d,k] + alpha) / (N - 1 - nb[d] + K * alpha) * \
                       (nwz[w,k] + beta) / (nz[k] + V * beta)
                # p(w|x = 1)
                s += p[k]

            # choice between p(x = 0,w) <-> p(x = 1,w)
            p0 = q * (nw[w] + eta) / (nws[0] + V * eta)
            p1 = (1 - q) * s

            # draw topic
            if (stopwords is not None) and (stopwords[w] > 0):
                z = 0
            elif (rands[n] < (p0 / (p0 + p1))):
                z = 0
            else:
                t = 0.0
                r = rands[N + n] * s
                for k in range(K):
                    t += p[k]
                    if (r < t):
                        z = 1 + k
                        break
            zz[d][n] = z
            
            # increment
            if z == 0:
                nb[d]    += 1
                # global statistics
                nw[w]    += 1
                nws[0]   += 1
            else:
                z = z - 1
                ndz[d,z] += 1
                nwz[w,z] += 1
                nz[z]    += 1
                
                
def infer (list words,
           np.ndarray[int,ndim=1] zz not None,
           np.ndarray[double,ndim=1] pzz not None,
           double alpha,
           np.ndarray[double,ndim=2] beta not None,
           np.ndarray[double,ndim=1] p0 not None,
           int iters):

    cdef int N = len(words)
    cdef int V = beta.shape[0]
    cdef int K = beta.shape[1]
    cdef np.ndarray[ftype_t,ndim=1] p = np.zeros (K, dtype=np.float64)
    cdef np.ndarray[dtype_t,ndim=1] nz = np.zeros (K, dtype=np.int32)
    cdef np.ndarray[dtype_t,ndim=1] seq = np.zeros (N, dtype=np.int32)
    cdef np.ndarray[ftype_t,ndim=1] rands
    cdef double a0 = 1.0, a1 = 1.0
    cdef int nb = 0, nsamples = 0
    # variables
    cdef int n, j, k, z
    cdef double q, s, t, r, r0, r1

    # prepare
    for n in range(N):
        seq[n] = n

    # sample
    for iter in range(iters):
        npr.shuffle (seq)
        rands = npr.rand (2 * N)
        for j in range(N):
            n = seq[j]
            w = words[n]
            z = zz[n]
            s = 0.0

            # decrement
            if (iter > 0):
                if z == 0:
                    nb -= 1
                else:
                    z = z - 1
                    nz[z] -= 1
            
            # compute p(x = 0)
            if (iter > 0):
                q = (a0 + nb) / (a0 + a1 + N - 1)
            else:
                q = 0.5

            # p(z = k,w|x = 1) = p(w|z = k)p(z = k)
            for k in range(K):
                p[k] = (nz[k] + alpha) / (N - nb - 1 + K * alpha) * beta[w][k]
                # p(w|x = 1)
                s += p[k]

            # choice between p(x = 0,w) <-> p(x = 1,w)
            r0 = q * p0[w]
            r1 = (1 - q) * s

            # sample and record
            if (rands[n] < (r0 / (r0 + r1))):
                z = 0
            else:
                t = 0.0
                r = rands[N + n] * s
                for k in range(K):
                    t += p[k]
                    if (r < t):
                        z = 1 + k
                        break
            zz[n] = z

            if (iter > iters / 2):
                if z > 0:
                    pzz[n] += 1.0

            # increment
            if z == 0:
                nb += 1
            else:
                z = z - 1
                nz[z] += 1
                
        if (iter > iters / 2):
            nsamples += 1
            
    for n in range(N):
        pzz[n] /= nsamples
